"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Color Difference Equations
""" 
from math import *

def delta_e_cie1976(color1, color2):
    """
    Calculates the Delta E (CIE1976) of two colors.
    """        
    # Color 1 
    L1 = float(color1.lab_l)
    a1 = float(color1.lab_a)
    b1 = float(color1.lab_b)
    # Color 2
    L2 = float(color2.lab_l)
    a2 = float(color2.lab_a)
    b2 = float(color2.lab_b)
    
    delta_L = pow(L1 - L2, 2)
    delta_a = pow(a1 - a2, 2)
    delta_b = pow(b1 - b2, 2)
    
    return sqrt(delta_L + delta_a + delta_b)

def delta_e_cie1994(color1, color2, K_L=1, K_C=1, K_H=1, K_1=0.045, K_2=0.015):
    """
    Calculates the Delta E (CIE1994) of two colors.
    
    K_l:
      0.045 graphic arts
      0.048 textiles
    K_2:
      0.015 graphic arts
      0.014 textiles
    K_L:
      1 default
      2 textiles
    """        
    # Color 1 
    L1 = float(color1.lab_l)
    a1 = float(color1.lab_a)
    b1 = float(color1.lab_b)
    # Color 2
    L2 = float(color2.lab_l)
    a2 = float(color2.lab_a)
    b2 = float(color2.lab_b)
        
    C_1 = sqrt(pow(a1, 2) + pow(b1, 2))
    C_2 = sqrt(pow(a2, 2) + pow(b2, 2))
    
    S_L = 1
    S_C = 1 + K_1 * C_1 
    S_H = 1 + K_2 * C_1

    delta_L = L1 - L2
    delta_C = C_1 - C_2
    delta_a = a1 - a2
    delta_b = b1 - b2
    
    try:
        delta_H = sqrt(pow(delta_a, 2) + pow(delta_b, 2) - pow(delta_C, 2))
    except ValueError:
        delta_H = 0.0
    
    L_group = pow(delta_L / (K_L * S_L), 2)
    C_group = pow(delta_C / (K_C * S_C), 2)
    H_group = pow(delta_H / (K_H * S_H), 2)
    
    return sqrt(L_group + C_group + H_group)

def delta_e_cie2000(color1, color2, Kl=1, Kc=1, Kh=1):
    """
    Calculates the Delta E (CIE2000) of two colors.
    """        
    # Color 1 
    L1 = float(color1.lab_l)
    a1 = float(color1.lab_a)
    b1 = float(color1.lab_b)
    # Color 2
    L2 = float(color2.lab_l)
    a2 = float(color2.lab_a)
    b2 = float(color2.lab_b)

    avg_Lp = (L1 + L2) / 2.0
    C1 = sqrt(pow(a1, 2) + pow(b1, 2))
    C2 = sqrt(pow(a2, 2) + pow(b2, 2))
    avg_C1_C2 = (C1 + C2) / 2.0 

    G = 0.5 * (1 - sqrt(pow(avg_C1_C2 , 7.0) / (pow(avg_C1_C2, 7.0) + pow(25.0, 7.0))))

    a1p = (1.0 + G) * a1
    a2p = (1.0 + G) * a2
    C1p = sqrt(pow(a1p, 2) + pow(b1, 2))
    C2p = sqrt(pow(a2p, 2) + pow(b2, 2))
    avg_C1p_C2p =(C1p + C2p) / 2.0
   
    if degrees(atan2(b1,a1p)) >= 0:
        h1p = degrees(atan2(b1,a1p))
    else:
        h1p = degrees(atan2(b1,a1p)) + 360
      
    if degrees(atan2(b2,a2p)) >= 0:
        h2p = degrees(atan2(b2,a2p))
    else:
        h2p = degrees(atan2(b2,a2p)) + 360
      
    if fabs(h1p - h2p) > 180:
        avg_Hp = (h1p + h2p + 360) / 2.0
    else:
        avg_Hp = (h1p + h2p) / 2.0

    T = 1 - 0.17 * cos(radians(avg_Hp - 30)) + 0.24 * cos(radians(2 * avg_Hp)) + 0.32 * cos(radians(3 * avg_Hp + 6)) - 0.2  * cos(radians(4 * avg_Hp - 63))

    diff_h2p_h1p = h2p - h1p
    if fabs(diff_h2p_h1p) <= 180:
        delta_hp = diff_h2p_h1p
    elif (fabs(diff_h2p_h1p) > 180) and (h2p <= h1p):
        delta_hp = diff_h2p_h1p + 360
    else:
        delta_hp = diff_h2p_h1p - 360
      
    delta_Lp = L2 - L1
    delta_Cp = C2p - C1p
    delta_Hp = 2 * sqrt(C2p * C1p) * sin(radians(delta_hp) / 2.0)
   
    S_L = 1 + ((0.015 * pow(avg_Lp - 50, 2)) / sqrt(20 + pow(avg_Lp - 50, 2.0)))
    S_C = 1 + 0.045 * avg_C1p_C2p
    S_H = 1 + 0.015 * avg_C1p_C2p * T
   
    delta_ro = 30 * exp(-(pow(((avg_Hp - 275) / 25), 2.0)))
    R_C = sqrt((pow(avg_C1p_C2p, 7.0)) / (pow(avg_C1p_C2p, 7.0) + pow(25.0, 7.0)));
    R_T = -2 * R_C * sin(2 * radians(delta_ro))

    delta_E = sqrt(pow(delta_Lp /(S_L * Kl), 2) + pow(delta_Cp /(S_C * Kc), 2) + pow(delta_Hp /(S_H * Kh), 2) + R_T * (delta_Cp /(S_C * Kc)) * (delta_Hp / (S_H * Kh)))

    return delta_E

def delta_e_cmc(color1, color2, p1=2, pc=1):
    """
    Calculates the Delta E (CIE1994) of two colors.
    
    CMC values
      Acceptability: p1=2, pc=1
      Perceptability: p1=1, pc=1
    """        
    # Color 1 
    L1 = float(color1.lab_l)
    a1 = float(color1.lab_a)
    b1 = float(color1.lab_b)
    # Color 2
    L2 = float(color2.lab_l)
    a2 = float(color2.lab_a)
    b2 = float(color2.lab_b)
    
    delta_L = L1 - L2
    delta_a = a1 - a2
    delta_b = b1 - b2
    
    C_1 = sqrt(pow(a1, 2) + pow(b1, 2))
    C_2 = sqrt(pow(a2, 2) + pow(b2, 2))
    
    H_1 = degrees(atan2(b1, a1))
    F = sqrt(pow(C_1, 4) / (pow(C_1, 4) + 1900.0))
    if 164 <= H_1 and H_1 <= 345:
        T = 0.56 + abs(0.2 * cos(radians(H_1 + 168)))
    else:
        T = 0.36 + abs(0.4 * cos(radians(H_1 + 35)))
        
    if L1 < 16:
        S_L = 0.511
    else:
        S_L = (0.040975 * L1) / (1 + 0.01765 * L1)
    S_C = ((0.0638 * C_1) / (1 + 0.0131 * C_1)) + 0.638
    S_H = S_C * (F * T + 1 - F)
    
    delta_L = L1 - L2
    delta_C = C_1 - C_2
    try:
        delta_H = sqrt(pow(delta_a, 2) + pow(delta_b, 2) - pow(delta_C, 2))
    except ValueError:
        delta_H = 0.0
    
    L_group = delta_L / (p1 * S_L)
    C_group = delta_C / (pc * S_C)
    H_group = delta_H / S_H
    
    return sqrt(pow(L_group, 2) + pow(C_group, 2) + pow(H_group, 2))


"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module shows you how to perform color space conversions. Please see the
chart on www.brucelindbloom.com/Math.html for an illustration of the conversions
you may perform.
"""
import example_config
from colormath.color_objects import LabColor, LCHabColor, SpectralColor, RGBColor

def example_lab_to_xyz():
    """
    This function shows a simple conversion of an Lab color to an XYZ color
    with debugging on (to show verbose output so you can see what the library
    is doing for you).
    """
    print "=== Simple Example: Lab->XYZ ==="
    # Instantiate an Lab color object with the given values.
    lab = LabColor(0.903, 16.296, -2.22)
    # Show a string representation.
    print lab
    # Convert to XYZ.
    xyz = lab.convert_to('xyz', debug=False)
    print xyz
    print "=== End Example ===\n"
    
def example_lchab_to_lchuv():
    """
    This function shows very complex chain of conversions in action. While this
    requires no additional effort from the user, debugging mode will show you
    all of the work that goes on behind the scenes.
    
    LCHab to LCHuv involves four different calculations, making this the
    conversion requiring the most steps.
    """
    print "=== Complex Example: LCHab->LCHuv ==="
    # Instantiate an LCHab color object with the given values.
    lchab = LCHabColor(0.903, 16.447, 352.252)
    # Show a string representation.
    print lchab
    # Convert to LCHuv.
    lchuv = lchab.convert_to('lchuv', debug=False)
    print lchuv
    print "=== End Example ===\n"
    
def example_lab_to_rgb():
    """
    Conversions to RGB are a little more complex mathematically. There are also
    several kinds of RGB color spaces. When converting from a device-independent
    color space to RGB, sRGB is assumed unless otherwise specified with the
    target_rgb keyword arg.
    """
    print "=== RGB Example: Lab->RGB ==="
    # Instantiate an Lab color object with the given values.
    lab = LabColor(0.903, 16.296, -2.217)
    # Show a string representation.
    print lab
    # Convert to XYZ.
    rgb = lab.convert_to('rgb', target_rgb='sRGB', debug=False)
    print rgb
    print "=== End Example ===\n"
    
def example_rgb_to_xyz():
    """
    The reverse is similar.
    """
    print "=== RGB Example: RGB->XYZ ==="
    # Instantiate an Lab color object with the given values.
    rgb = RGBColor(120, 130, 140, rgb_type='sRGB')
    # Show a string representation.
    print rgb
    # Convert RGB to XYZ using a D50 illuminant.
    xyz = rgb.convert_to('xyz', target_illuminant='D50')
    print xyz
    print "=== End Example ===\n"
    
def example_spectral_to_xyz():
    print "=== Example: Spectral->XYZ ==="
    """
    Instantiate an Lab color object with the given values. Note that the
    spectral range can run from 340nm to 830nm. Any omitted values assume a
    value of 0.0, which is more or less ignored. For the distribution below,
    we are providing an example reading from an X-Rite i1 Pro, which only
    measures between 380nm and 730nm.
    """
    spc = SpectralColor(observer=2, illuminant='d50', 
                        spec_380nm=0.0600, spec_390nm=0.0600, spec_400nm=0.0641,
                        spec_410nm=0.0654, spec_420nm=0.0645, spec_430nm=0.0605,
                        spec_440nm=0.0562, spec_450nm=0.0543, spec_460nm=0.0537,
                        spec_470nm=0.0541, spec_480nm=0.0559, spec_490nm=0.0603,
                        spec_500nm=0.0651, spec_510nm=0.0680, spec_520nm=0.0705,
                        spec_530nm=0.0736, spec_540nm=0.0772, spec_550nm=0.0809,
                        spec_560nm=0.0870, spec_570nm=0.0990, spec_580nm=0.1128,
                        spec_590nm=0.1251, spec_600nm=0.1360, spec_610nm=0.1439,
                        spec_620nm=0.1511, spec_630nm=0.1590, spec_640nm=0.1688,
                        spec_650nm=0.1828, spec_660nm=0.1996, spec_670nm=0.2187,
                        spec_680nm=0.2397, spec_690nm=0.2618, spec_700nm=0.2852,
                        spec_710nm=0.2500, spec_720nm=0.2400, spec_730nm=0.2300)
    xyz = spc.convert_to('xyz')
    print xyz
    print "=== End Example ===\n"
    
# Feel free to comment/un-comment examples as you please.
example_lab_to_xyz()
example_lchab_to_lchuv()
example_lab_to_rgb()
example_spectral_to_xyz()
example_rgb_to_xyz()


"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module shows you how to perform various kinds of density calculations.
"""
import example_config
from colormath.color_objects import SpectralColor
from colormath.density_standards import *

EXAMPLE_COLOR = SpectralColor(observer=2, illuminant='d50', 
                    spec_380nm=0.0600, spec_390nm=0.0600, spec_400nm=0.0641,
                    spec_410nm=0.0654, spec_420nm=0.0645, spec_430nm=0.0605,
                    spec_440nm=0.0562, spec_450nm=0.0543, spec_460nm=0.0537,
                    spec_470nm=0.0541, spec_480nm=0.0559, spec_490nm=0.0603,
                    spec_500nm=0.0651, spec_510nm=0.0680, spec_520nm=0.0705,
                    spec_530nm=0.0736, spec_540nm=0.0772, spec_550nm=0.0809,
                    spec_560nm=0.0870, spec_570nm=0.0990, spec_580nm=0.1128,
                    spec_590nm=0.1251, spec_600nm=0.1360, spec_610nm=0.1439,
                    spec_620nm=0.1511, spec_630nm=0.1590, spec_640nm=0.1688,
                    spec_650nm=0.1828, spec_660nm=0.1996, spec_670nm=0.2187,
                    spec_680nm=0.2397, spec_690nm=0.2618, spec_700nm=0.2852,
                    spec_710nm=0.2500, spec_720nm=0.2400, spec_730nm=0.2300)

def example_auto_status_t_density():
    print "=== Example: Automatic Status T Density ==="
    # If no arguments are provided to calc_density(), ANSI Status T density is
    # assumed. The correct RGB "filter" is automatically selected for you.
    print "Density: %f" % EXAMPLE_COLOR.calc_density()
    print "=== End Example ===\n"
    
def example_manual_status_t_density():
    print "=== Example: Manual Status T Density ==="
    # Here we are specifically requesting the value of the red band via the
    # ANSI Status T spec.
    print "Density: %f (Red)" % EXAMPLE_COLOR.calc_density(density_standard=ANSI_STATUS_T_RED)
    print "=== End Example ===\n"
    
def example_visual_density():
    print "=== Example: Visual Density ==="
    # Here we pass the ISO Visual spectral standard.
    print "Density: %f" % EXAMPLE_COLOR.calc_density(density_standard=ISO_VISUAL)
    print "=== End Example ===\n"
    
# Feel free to comment/un-comment examples as you please.
example_auto_status_t_density()
example_manual_status_t_density()
example_visual_density()



"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module contains classes to represent various color spaces.
"""
import numpy
from colormath import color_conversions
from colormath import color_constants
from colormath import density
from colormath.color_exceptions import *
from colormath.color_diff import delta_e_cie2000, delta_e_cie1976, delta_e_cie1994, delta_e_cmc

class ColorBase(object):
    """
    A base class holding some common methods and values.
    """
    def __init__(self, *args, **kwargs):
        # This is the most common illuminant, default to it.
        self.illuminant = 'd50'
        # This is the most commonly used observer angle.
        self.observer = '2'
        self.OTHER_VALUES = ['illuminant', 'observer']
        
    def _transfer_kwargs(self, *args, **kwargs):
        """
        Transfers any keyword arguments to the appropriate coordinate fields
        if they match one of the keys in the class's VALUES dict.
        
        Also transfers *args to the corresponding tristimulus values.
        """
        # Used for tracking which member of the VALUES list we're on.
        counter = 0
        # This is the max number of args used for VALUES.
        max_args = len(self.VALUES) - 1
        for arg in args:
            if counter <= max_args:
                # Transfer the tri-stim value.
                setattr(self, self.VALUES[counter], arg)
                counter += 1

        # Tranfser matching keywords.
        attrib_list = self.VALUES + self.OTHER_VALUES
        for key, val in kwargs.items():
            if key in attrib_list:
                # This is useful for stuff like illuminants. All of our
                # constants dictionaries are lowercase.
                if isinstance(val, str):
                    val = val.lower()
                # Transfer the value to the color object.
                setattr(self, key, val)
                
    def __prep_strings(self):
        """
        Makes sure all string variables are lowercase beforehand.
        """
        self.illuminant = self.illuminant.lower()
        self.observer = str(self.observer)

    def convert_to(self, cs_to, *args, **kwargs):
        """
        Converts the color to the designated colorspace.
        """
        debug = kwargs.get('debug', False)
        try:
            # Look up the conversion path for the specified color space.
            conversions = self.CONVERSIONS[cs_to.lower()]
        except KeyError:
            raise InvalidConversion(self.__class__.__name__, cs_to)
        
        # Make sure any string variables are lowercase.
        self.__prep_strings()
        # Make sure the object has all of its required values before even
        # attempting a conversion.
        self.has_required_values()
        
        if debug:
            print 'Converting %s to %s' % (self, cs_to)
            print ' @ Conversion path: %s' % [conv.__name__ for conv in conversions]

        cobj = self
        # Iterate through the list of functions for the conversion path, storing
        # the results in a dictionary via update(). This way the user has access
        # to all of the variables involved in the conversion.
        for func in conversions:
            # Execute the function in this conversion step and store the resulting
            # Color object.
            if debug:
                print ' * Conversion: %s passed to %s()' % (
                                        cobj.__class__.__name__, func.__name__)
                print ' |->  in %s' % cobj
                
            if func:
                # This can be None if you try to convert a color to the color
                # space that is already in. IE: XYZ->XYZ.
                cobj = func(cobj, *args, **kwargs)
            
            if debug:
                print ' |-< out %s' % cobj
        return cobj
    
    def get_value_tuple(self):
        """
        Returns a tuple of the color's values (in order). For example,
        an LabColor object will return (lab_l, lab_a, lab_b), where each
        member of the tuple is the float value for said variable.
        """
        retval = tuple()
        for val in self.VALUES:
            retval += (getattr(self, val, None),)
        return retval
    
    def __str__(self):
        """
        String representation of the color.
        """
        retval = self.__class__.__name__ + ' ('
        for val in self.VALUES:
            value = getattr(self, val, None)
            #print "VAL: %s Type: %s" % (val, value)
            if value != None:
                retval += '%s:%.4f ' % (val, getattr(self, val))
        return retval.strip() + ')'
    
    def has_required_values(self):
        """
        Checks various fields for None or invalid values.
        """
        if self.observer not in ['2', '10']:
            raise InvalidObserver(self)
        
        for val in self.VALUES:
            value = getattr(self, val, None)
            if value == None:
                # A required value is missing.
                raise MissingValue(self, val)
            
            try:
                # If this fails, it's not a usable number.
                float(value)
            except ValueError:
                raise InvalidValue(self, val, value)
        return True
    
    def get_illuminant_xyz(self, observer=None, illuminant=None):
        """
        Returns the color's illuminant's XYZ values.
        
        observer: (string) Get the XYZ values for another observer angle. Must
                           be either '2' or '10'.
        illuminant: (string) Get the XYZ values for another illuminant.
        """
        try:
            if observer == None:
                observer = self.observer
                
            illums_observer = color_constants.ILLUMINANTS[self.observer]
        except KeyError:
            raise InvalidObserver(self)
        
        try:
            if illuminant == None:
                illuminant = self.illuminant
                
            illum_xyz = illums_observer[illuminant]
        except AttributeError:
            raise InvalidIlluminant(self)
        except KeyError:
            raise InvalidIlluminant(self)
        
        return {'X': illum_xyz[0], 'Y': illum_xyz[1], 'Z': illum_xyz[2]}
    
    def delta_e(self, other_color, mode='cie2000', *args, **kwargs):
        """
        Compares this color to another via Delta E.
        
        Valid modes:
         cie2000
         cie1976
        """
        if not isinstance(other_color, ColorBase):
            raise InvalidArgument('delta_e_cie2000', 'other_color', other_color)
        
        # Convert the colors to Lab if they are not already.
        lab1 = self.convert_to('lab', *args, **kwargs)
        lab2 = other_color.convert_to('lab', *args, **kwargs)
        
        mode = mode.lower()
        if mode == 'cie2000':
            return delta_e_cie2000(lab1, lab2)
        elif mode == 'cie1994':
            return delta_e_cie1994(lab1, lab2, **kwargs)
        elif mode == 'cie1976':
            return delta_e_cie1976(lab1, lab2)
        elif mode == 'cmc':
            return delta_e_cmc(lab1, lab2, **kwargs)
        else:
            raise InvalidDeltaEMode(mode)

class SpectralColor(ColorBase):
    """
    Represents a color that may have operations done to it. You need not use
    this object with the library as long as you use all of the instance
    variables here.
    """
    CONVERSIONS = {
        "spectral": [None],
        "xyz": [color_conversions.Spectral_to_XYZ],
        "xyy": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_Lab],
        "lch": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
        "luv": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_Luv],
        "rgb": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.Spectral_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY, color_conversions.CMY_to_CMYK],
    }
    VALUES = ['spec_340nm', 'spec_350nm', 'spec_360nm', 'spec_370nm',
              'spec_380nm', 'spec_390nm', 'spec_400nm', 'spec_410nm', 
              'spec_420nm', 'spec_430nm', 'spec_440nm', 'spec_450nm', 
              'spec_460nm', 'spec_470nm', 'spec_480nm', 'spec_490nm', 
              'spec_500nm', 'spec_510nm', 'spec_520nm', 'spec_530nm', 
              'spec_540nm', 'spec_550nm', 'spec_560nm', 'spec_570nm', 
              'spec_580nm', 'spec_590nm', 'spec_600nm', 'spec_610nm', 
              'spec_620nm', 'spec_630nm', 'spec_640nm', 'spec_650nm', 
              'spec_660nm', 'spec_670nm', 'spec_680nm', 'spec_690nm', 
              'spec_700nm', 'spec_710nm', 'spec_720nm', 'spec_730nm',
              'spec_740nm', 'spec_750nm', 'spec_760nm', 'spec_770nm',
              'spec_780nm', 'spec_790nm', 'spec_800nm', 'spec_810nm',
              'spec_820nm', 'spec_830nm']
    
    def __init__(self, *args, **kwargs):
        super(SpectralColor, self).__init__(*args, **kwargs)
        # Spectral fields
        self.spec_340nm = 0.0
        self.spec_350nm = 0.0
        self.spec_360nm = 0.0
        self.spec_370nm = 0.0
        self.spec_380nm = 0.0 # begin Blue wavelengths
        self.spec_390nm = 0.0
        self.spec_400nm = 0.0
        self.spec_410nm = 0.0
        self.spec_420nm = 0.0
        self.spec_430nm = 0.0
        self.spec_440nm = 0.0
        self.spec_450nm = 0.0
        self.spec_460nm = 0.0
        self.spec_470nm = 0.0
        self.spec_480nm = 0.0
        self.spec_490nm = 0.0 # end Blue wavelengths
        self.spec_500nm = 0.0 # start Green wavelengths
        self.spec_510nm = 0.0
        self.spec_520nm = 0.0
        self.spec_530nm = 0.0
        self.spec_540nm = 0.0
        self.spec_550nm = 0.0
        self.spec_560nm = 0.0
        self.spec_570nm = 0.0
        self.spec_580nm = 0.0
        self.spec_590nm = 0.0
        self.spec_600nm = 0.0
        self.spec_610nm = 0.0 # end Green wavelengths
        self.spec_620nm = 0.0 # start Red wavelengths
        self.spec_630nm = 0.0
        self.spec_640nm = 0.0
        self.spec_650nm = 0.0
        self.spec_660nm = 0.0
        self.spec_670nm = 0.0
        self.spec_680nm = 0.0
        self.spec_690nm = 0.0
        self.spec_700nm = 0.0
        self.spec_710nm = 0.0
        self.spec_720nm = 0.0
        self.spec_730nm = 0.0 # end Red wavelengths
        self.spec_740nm = 0.0
        self.spec_750nm = 0.0
        self.spec_760nm = 0.0
        self.spec_770nm = 0.0
        self.spec_780nm = 0.0
        self.spec_790nm = 0.0
        self.spec_800nm = 0.0
        self.spec_810nm = 0.0
        self.spec_820nm = 0.0
        self.spec_830nm = 0.0
        self._transfer_kwargs(*args, **kwargs)
        
    def get_numpy_array(self):
        """
        Dump this color into NumPy array.
        """
        # This holds the obect's spectral data, and will be passed to 
        # numpy.array() to create a numpy array (matrix) for the matrix math
        # that will be done during the conversion to XYZ. 
        values = []
        
        # Use the required value list to build this dynamically. Default to
        # 0.0, since that ultimately won't affect the outcome due to the math
        # involved.
        for val in self.VALUES:
            values.append(getattr(self, val, 0.0))

        # Create and the actual numpy array/matrix from the spectral list.
        color_array = numpy.array([values])
        return color_array
    
    def calc_density(self, density_standard=None):
        """
        Calculates the density of the SpectralColor. By default, Status T
        density is used, and the correct density distribution (Red, Green,
        or Blue) is chosen by comparing the Red, Green, and Blue components of
        the spectral sample (the values being red in via "filters").
        """
        if density_standard != None:
            return density.ansi_density(self, density_standard)
        else:
            return density.auto_density(self)
    
class LabColor(ColorBase):
    """
    Represents an Lab color.
    """
    CONVERSIONS = {
        "lab": [None],
        "xyz": [color_conversions.Lab_to_XYZ],
        "xyy": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_xyY],
      "lchab": [color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_Luv,
                color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_Luv],
        "rgb": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.Lab_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY, color_conversions.CMY_to_CMYK],
    }
    VALUES = ['lab_l', 'lab_a', 'lab_b']
       
    def __init__(self, *args, **kwargs):
        super(LabColor, self).__init__(*args, **kwargs)
        self.lab_l = None
        self.lab_a = None
        self.lab_b = None
        self._transfer_kwargs(*args, **kwargs)
        
class LCHabColor(ColorBase):
    """
    Represents an LCHab color.
    """
    CONVERSIONS = {
      "lchab": [None],
        "xyz": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ],
        "xyy": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ, 
                color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.LCHab_to_Lab],
      "lchuv": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ,
                color_conversions.XYZ_to_Luv, color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ, 
                color_conversions.XYZ_to_Luv],
        "rgb": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ, 
                color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ, 
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ, 
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ,
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.LCHab_to_Lab, color_conversions.Lab_to_XYZ,
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY, 
                color_conversions.CMY_to_CMYK],
    }
    VALUES = ['lch_l', 'lch_c', 'lch_h']
    
    def __init__(self, *args, **kwargs):
        super(LCHabColor, self).__init__(*args, **kwargs)
        self.lch_l = None
        self.lch_c = None
        self.lch_h = None
        self._transfer_kwargs(*args, **kwargs)
        
class LCHuvColor(ColorBase):
    """
    Represents an LCHuv color.
    """
    CONVERSIONS = {
      "lchuv": [None],
        "xyz": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ],
        "xyy": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ, 
                color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ, 
                color_conversions.XYZ_to_Lab],
        "luv": [color_conversions.LCHuv_to_Luv],
      "lchab": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ,
                color_conversions.XYZ_to_Lab, color_conversions.Lab_to_LCHab],
        "rgb": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ, 
                color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ, 
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ, 
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ,
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.LCHuv_to_Luv, color_conversions.Luv_to_XYZ,
                color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY, 
                color_conversions.CMY_to_CMYK],
    }
    VALUES = ['lch_l', 'lch_c', 'lch_h']
    
    def __init__(self, *args, **kwargs):
        super(LCHuvColor, self).__init__(*args, **kwargs)
        self.lch_l = None
        self.lch_c = None
        self.lch_h = None
        self._transfer_kwargs(*args, **kwargs)
        
class LuvColor(ColorBase):
    """
    Represents an Luv color.
    """
    CONVERSIONS = {
        "luv": [None],
        "xyz": [color_conversions.Luv_to_XYZ],
        "xyy": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.Luv_to_LCHuv],
        "rgb": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.Luv_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_CMY, color_conversions.CMY_to_CMYK],
    }
    VALUES = ['luv_l', 'luv_u', 'luv_v']
    
    def __init__(self, *args, **kwargs):
        super(LuvColor, self).__init__(*args, **kwargs)
        self.luv_l = None
        self.luv_u = None
        self.luv_v = None
        self._transfer_kwargs(*args, **kwargs)
        
class XYZColor(ColorBase):
    """
    Represents an XYZ color.
    """
    CONVERSIONS = {
        "xyz": [None],
        "xyy": [color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.XYZ_to_Lab, color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.XYZ_to_Lab, color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.XYZ_to_Luv],
        "rgb": [color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.XYZ_to_RGB, color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.XYZ_to_RGB, color_conversions.RGB_to_CMY, 
                color_conversions.CMY_to_CMYK],
    }
    VALUES = ['xyz_x', 'xyz_y', 'xyz_z']
    
    def apply_adaptation(self, target_illuminant, adaptation='bradford', 
                         debug=False):
        """
        This applies an adaptation matrix to change the XYZ color's illuminant. 
        You'll most likely only need this during RGB conversions.
        """            
        # The illuminant of the original RGB object.
        source_illuminant = self.illuminant
        
        if debug:
            print "  \- Original illuminant: %s" % self.illuminant
            print "  \- Target illuminant: %s" % target_illuminant
       
        # If the XYZ values were taken with a different reference white than the
        # native reference white of the target RGB space, a transformation matrix
        # must be applied.
        if source_illuminant != target_illuminant:
            if debug:
                print "  \* Applying transformation from %s to %s " % (
                                                            source_illuminant,
                                                            target_illuminant)
            # Get the adjusted XYZ values, adapted for the target illuminant.
            self.xyz_x, self.xyz_y, self.xyz_z = color_conversions.apply_XYZ_transformation(
                                                   self.xyz_x,
                                                   self.xyz_y,
                                                   self.xyz_z, 
                                                   orig_illum=source_illuminant, 
                                                   targ_illum=target_illuminant,
                                                   debug=debug)
            self.illuminant = target_illuminant.lower()
    
    def __init__(self, *args, **kwargs):
        super(XYZColor, self).__init__(*args, **kwargs)
        self.xyz_x = None
        self.xyz_y = None
        self.xyz_z = None
        self._transfer_kwargs(*args, **kwargs)
        
class xyYColor(ColorBase):
    """
    Represents an xYy color.
    """
    CONVERSIONS = {
        "xyy": [None],
        "xyz": [color_conversions.xyY_to_XYZ],
        "lab": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_Luv, 
                color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_Luv],
        "rgb": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_RGB],
        "hsl": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_RGB, 
                color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_RGB,
                color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.xyY_to_XYZ, color_conversions.XYZ_to_RGB,
                color_conversions.RGB_to_CMY, color_conversions.CMY_to_CMYK],
    }
    VALUES = ['xyy_x', 'xyy_y', 'xyy_Y']
    
    def __init__(self, *args, **kwargs):
        super(xyYColor, self).__init__(*args, **kwargs)
        self.xyy_x = None
        self.xyy_y = None
        self.xyy_Y = None
        self._transfer_kwargs(*args, **kwargs)
        
class RGBColor(ColorBase):
    """
    Represents an Lab color.
    """
    CONVERSIONS = {
        "rgb": [None],
        "hsl": [color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.RGB_to_HSV],
        "cmy": [color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.RGB_to_CMY, color_conversions.CMY_to_CMYK],
        "xyz": [color_conversions.RGB_to_XYZ],
        "xyy": [color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Luv, 
                color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Luv],
    }
    VALUES = ['rgb_r', 'rgb_g', 'rgb_b']
    
    def __init__(self, *args, **kwargs):
        super(RGBColor, self).__init__(*args, **kwargs)
        self.rgb_r = None
        self.rgb_g = None
        self.rgb_b = None
        self.rgb_type = 'srgb'
        self.OTHER_VALUES.append('rgb_type')
        self._transfer_kwargs(*args, **kwargs)
        
    def __str__(self):
        parent_str = super(RGBColor, self).__str__()
        return '%s [%s]' % (parent_str, self.rgb_type)
    
    def get_rgb_hex(self):
        """
        Converts the RGB value to a hex value in the form of: #RRGGBB
        """
        self.has_required_values()
        return '#%02x%02x%02x' % (self.rgb_r, self.rgb_g, self.rgb_b)
    
    def set_from_rgb_hex(self, hex_str):
        """
        Converts an RGB hex string like #RRGGBB and assigns the values to
        this RGBColor object.
        """
        colorstring = hex_str.strip()
        if colorstring[0] == '#': 
            colorstring = colorstring[1:]
        if len(colorstring) != 6:
            raise ValueError, "input #%s is not in #RRGGBB format" % colorstring
        r, g, b = colorstring[:2], colorstring[2:4], colorstring[4:]
        r, g, b = [int(n, 16) for n in (r, g, b)]
        self.rgb_r = r
        self.rgb_g = g
        self.rgb_b = b
        
class HSLColor(ColorBase):
    """
    Represents a HSL color.
    """
    CONVERSIONS = {
        "hsl": [None],
        "hsv": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_HSV],
        "rgb": [color_conversions.HSL_to_RGB],
        "cmy": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_CMY,
                color_conversions.CMY_to_CMYK],
        "xyz": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ],
        "xyy": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Lab, color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Luv, color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.HSL_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_RGB],
    }
    VALUES = ['hsl_h', 'hsl_s', 'hsl_l']
    
    def __init__(self, *args, **kwargs):
        super(HSLColor, self).__init__(*args, **kwargs)
        self.hsl_h = None
        self.hsl_s = None
        self.hsl_l = None
        self.rgb_type = 'srgb'
        self.OTHER_VALUES.append('rgb_type')
        self._transfer_kwargs(*args, **kwargs)
        
class HSVColor(ColorBase):
    """
    Represents an HSV color.
    """
    CONVERSIONS = {
        "hsv": [None],
        "hsl": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_HSL],
        "rgb": [color_conversions.HSV_to_RGB],
        "cmy": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_CMY],
       "cmyk": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_CMY,
                color_conversions.CMY_to_CMYK],
        "xyz": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ],
        "xyy": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Lab, color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_Luv, color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.HSV_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_RGB],
    }
    VALUES = ['hsv_h', 'hsv_s', 'hsv_v']
    
    def __init__(self, *args, **kwargs):
        super(HSVColor, self).__init__(*args, **kwargs)
        self.hsv_h = None
        self.hsv_s = None
        self.hsv_v = None
        self.rgb_type = 'srgb'
        self.OTHER_VALUES.append('rgb_type')
        self._transfer_kwargs(*args, **kwargs)
        
class CMYColor(ColorBase):
    """
    Represents a CMY color.
    """
    CONVERSIONS = {
        "cmy": [None],
       "cmyk": [color_conversions.CMY_to_CMYK],
        "hsl": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_HSV],
        "rgb": [color_conversions.CMY_to_RGB],
        "xyz": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ],
        "xyy": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ,
                color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_Luv, color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.CMY_to_RGB, color_conversions.RGB_to_XYZ, 
                color_conversions.XYZ_to_RGB],
    }
    VALUES = ['cmy_c', 'cmy_m', 'cmy_y']
    
    def __init__(self, *args, **kwargs):
        super(CMYColor, self).__init__(*args, **kwargs)
        self.cmy_c = None
        self.cmy_m = None
        self.cmy_y = None
        self._transfer_kwargs(*args, **kwargs)
        
class CMYKColor(ColorBase):
    """
    Represents a CMYK color.
    """
    CONVERSIONS = {
       "cmyk": [None],
        "cmy": [color_conversions.CMYK_to_CMY],
        "hsl": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_HSL],
        "hsv": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB,
                color_conversions.RGB_to_HSV],
        "rgb": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB],
        "xyz": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ],
        "xyy": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_xyY],
        "lab": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Lab],
      "lchab": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Lab, 
                color_conversions.Lab_to_LCHab],
      "lchuv": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_Luv, 
                color_conversions.Luv_to_LCHuv],
        "luv": [color_conversions.CMYK_to_CMY, color_conversions.CMY_to_RGB, 
                color_conversions.RGB_to_XYZ, color_conversions.XYZ_to_RGB],
    }
    VALUES = ['cmyk_c', 'cmyk_m', 'cmyk_y', 'cmyk_k']
    
    def __init__(self, *args, **kwargs):
        super(CMYKColor, self).__init__(*args, **kwargs)
        self.cmyk_c = None
        self.cmyk_m = None
        self.cmyk_y = None
        self.cmyk_k = None
        self._transfer_kwargs(*args, **kwargs)


"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Conversion between color spaces
"""
import math
import numpy
from numpy.linalg import pinv
from colormath import color_constants
from colormath import spectral_constants
from colormath.color_exceptions import InvalidIlluminant

def _transfer_common(old_cobj, new_cobj):
    """
    Transfer illuminant and observer data to a new color object. This is
    """
    new_cobj.illuminant = old_cobj.illuminant
    new_cobj.observer = old_cobj.observer

def _get_adaptation_matrix(orig_illum, targ_illum, observer, adaptation):
    """
    Calculate the correct transformation matrix based on origin and target
    illuminants. The observer angle must be the same between illuminants.
    
    See colormath.color_constants.ADAPTATION_MATRICES for a list of possible
    adaptations.
    
    Detailed conversion documentation is available at:
    http://brucelindbloom.com/Eqn_ChromAdapt.html
    """
    # Get the appropriate transformation matrix, [MsubA].
    transform_matrix = color_constants.ADAPTATION_MATRICES[adaptation]
    # Calculate the inverse of the transform matrix, [MsubA]^(-1)
    transform_matrix_inverse = pinv(transform_matrix)
    
    # Store the XYZ coordinates of the origin illuminant. Becomes XsubWS.
    illum_from = color_constants.ILLUMINANTS['2'][orig_illum]
    # Also store the XYZ coordinates of the target illuminant. Becomes XsubWD.
    illum_to = color_constants.ILLUMINANTS['2'][targ_illum]
    
    # Calculate cone response domains.
    pyb_source = numpy.dot(illum_from, transform_matrix)
    pyb_dest = numpy.dot(illum_to, transform_matrix)
    
    # Break the cone response domains out into their appropriate variables.
    P_sub_S, Y_sub_S, B_sub_S = pyb_source[0], pyb_source[1], pyb_source[2]
    P_sub_D, Y_sub_D, B_sub_D = pyb_dest[0], pyb_dest[1], pyb_dest[2]
    
    # Assemble the middle matrix used in the final calculation of [M].
    middle_matrix = numpy.array(((P_sub_D / P_sub_S, 0.0, 0.0),
                                 (0.0, Y_sub_D / Y_sub_S, 0.0),
                                 (0.0, 0.0, B_sub_D / B_sub_S)))
    
    return numpy.dot(numpy.dot(transform_matrix, middle_matrix), 
                  transform_matrix_inverse)

def apply_XYZ_transformation(val_x, val_y, val_z, orig_illum, targ_illum,
                             observer='2', adaptation='bradford', debug=False):
    """
    Applies an XYZ transformation matrix to convert XYZ values between
    illuminants. It is important to recognize that color transformation results
    in color errors, determined by how far the original illuminant is from the
    target illuminant. For example, D65 to A could result in very high maximum
    deviances.
    
    An informative article with estimate average Delta E values for each
    illuminant conversion may be found at:
    
    http://brucelindbloom.com/ChromAdaptEval.html
    """
    # It's silly to have to do this, but some people may want to call this
    # function directly, so we'll protect them from messing up upper/lower case.
    orig_illum = orig_illum.lower()
    targ_illum = targ_illum.lower()
    adaptation = adaptation.lower()
   
    if debug:
        print "  \* Applying adaptation matrix: %s" % adaptation
    # Retrieve the appropriate transformation matrix from the constants.
    transform_matrix = _get_adaptation_matrix(orig_illum, targ_illum, 
                                              observer, adaptation)

    # Stuff the XYZ values into a NumPy matrix for conversion.
    XYZ_matrix = numpy.array((
        val_x, val_y, val_z
     ))
    # Perform the adaptation via matrix multiplication.
    result_matrix = numpy.dot(XYZ_matrix, transform_matrix)
    
    # Return individual X, Y, and Z coordinates.
    return result_matrix[0], result_matrix[1], result_matrix[2]

def apply_RGB_matrix(var1, var2, var3, rgb_type, convtype="xyz_to_rgb", 
                     debug=False):
    """
    Applies an RGB working matrix to convert from XYZ to RGB.
    The arguments are tersely named var1, var2, and var3 to allow for the passing
    of XYZ _or_ RGB values. var1 is X for XYZ, and R for RGB. var2 and var3
    follow suite.
    """
    rgb_type = rgb_type.lower()
    convtype = convtype.lower()
    # Retrieve the appropriate transformation matrix from the constants.
    rgb_matrix = color_constants.RGB_SPECS[rgb_type]["conversions"][convtype]
   
    if debug:
        print "  \* Applying RGB conversion matrix: %s->%s" % (rgb_type, convtype)
    # Stuff the RGB/XYZ values into a NumPy matrix for conversion.
    var_matrix = numpy.array((
        var1, var2, var3
    ))
    # Perform the adaptation via matrix multiplication.
    result_matrix = numpy.dot(var_matrix, rgb_matrix)
    return result_matrix[0], result_matrix[1], result_matrix[2]

def Spectral_to_XYZ(cobj, debug=False, illuminant_override=None, *args, **kwargs):
    """
    Converts spectral readings to XYZ.
    """
    xyzcolor = color_objects.XYZColor()
    _transfer_common(cobj, xyzcolor)
    
    # If the user provides an illuminant_override numpy array, use it.
    if illuminant_override:
        reference_illum = illuminant_override
    else:
        # Otherwise, look up the illuminant from known standards based
        # on the value of 'illuminant' pulled from the SpectralColor object.
        try:
            reference_illum = spectral_constants.REF_ILLUM_TABLE[cobj.illuminant]
        except KeyError:
            raise InvalidIlluminant(cobj)
        
    # Get the spectral distribution of the selected standard observer.
    if cobj.observer == '10':
        std_obs_x = spectral_constants.STDOBSERV_X10
        std_obs_y = spectral_constants.STDOBSERV_Y10
        std_obs_z = spectral_constants.STDOBSERV_Z10
    else:
        # Assume 2 degree, since it is theoretically the only other possibility.
        std_obs_x = spectral_constants.STDOBSERV_X2
        std_obs_y = spectral_constants.STDOBSERV_Y2
        std_obs_z = spectral_constants.STDOBSERV_Z2
     
    # This is a NumPy array containing the spectral distribution of the color.
    sample = cobj.get_numpy_array()
    
    # The denominator is constant throughout the entire calculation for X,
    # Y, and Z coordinates. Calculate it once and re-use.
    denom = std_obs_y * reference_illum
    
    # This is also a common element in the calculation whereby the sample
    # NumPy array is multiplied by the reference illuminant's power distribution
    # (which is also a NumPy array).
    sample_by_ref_illum = sample * reference_illum
        
    # Calculate the numerator of the equation to find X.
    x_numerator = sample_by_ref_illum * std_obs_x
    y_numerator = sample_by_ref_illum * std_obs_y
    z_numerator = sample_by_ref_illum * std_obs_z
    
    xyzcolor.xyz_x = x_numerator.sum() / denom.sum()
    xyzcolor.xyz_y = y_numerator.sum() / denom.sum()
    xyzcolor.xyz_z = z_numerator.sum() / denom.sum()
    
    return xyzcolor

def Lab_to_LCHab(cobj, debug=False, *args, **kwargs):
    """
    Convert from CIE Lab to LCH(ab).
    """
    lchcolor = color_objects.LCHabColor()
    _transfer_common(cobj, lchcolor)
   
    lchcolor.lch_l = cobj.lab_l
    lchcolor.lch_c = math.sqrt(math.pow(float(cobj.lab_a),2) + math.pow(float(cobj.lab_b),2))
    lchcolor.lch_h = math.atan2(float(cobj.lab_b), float(cobj.lab_a))
   
    if (lchcolor.lch_h > 0):
        lchcolor.lch_h = (lchcolor.lch_h / math.pi) * 180
    else:
        lchcolor.lch_h = 360 - (math.fabs(lchcolor.lch_h) / math.pi) * 180
      
    return lchcolor

def Lab_to_XYZ(cobj, debug=False, *args, **kwargs):
    """
    Convert from Lab to XYZ
    """
    illum = cobj.get_illuminant_xyz()
    xyzcolor = color_objects.XYZColor()
    _transfer_common(cobj, xyzcolor)
   
    xyzcolor.xyz_y = (cobj.lab_l + 16.0) / 116.0
    xyzcolor.xyz_x = cobj.lab_a / 500.0 + xyzcolor.xyz_y
    xyzcolor.xyz_z = xyzcolor.xyz_y - cobj.lab_b / 200.0
   
    if math.pow(xyzcolor.xyz_y, 3) > color_constants.CIE_E:
        xyzcolor.xyz_y = math.pow(xyzcolor.xyz_y, 3)
    else:
        xyzcolor.xyz_y = (xyzcolor.xyz_y - 16.0 / 116.0) / 7.787

    if math.pow(xyzcolor.xyz_x, 3) > color_constants.CIE_E:
        xyzcolor.xyz_x = math.pow(xyzcolor.xyz_x, 3)
    else:
        xyzcolor.xyz_x = (xyzcolor.xyz_x - 16.0 / 116.0) / 7.787
      
    if math.pow(xyzcolor.xyz_z, 3) > color_constants.CIE_E:
        xyzcolor.xyz_z = math.pow(xyzcolor.xyz_z, 3)
    else:
        xyzcolor.xyz_z = (xyzcolor.xyz_z - 16.0 / 116.0) / 7.787
      
    xyzcolor.xyz_x = (illum["X"] * xyzcolor.xyz_x)
    xyzcolor.xyz_y = (illum["Y"] * xyzcolor.xyz_y)
    xyzcolor.xyz_z = (illum["Z"] * xyzcolor.xyz_z)
    
    return xyzcolor

def Luv_to_LCHuv(cobj, debug=False, *args, **kwargs):
    """
    Convert from CIE Luv to LCH(uv).
    """
    lchcolor = color_objects.LCHuvColor()
    _transfer_common(cobj, lchcolor)
   
    lchcolor.lch_l = cobj.luv_l
    lchcolor.lch_c = math.sqrt(math.pow(cobj.luv_u, 2.0) + math.pow(cobj.luv_v, 2.0))
    lchcolor.lch_h = math.atan2(float(cobj.luv_v), float(cobj.luv_u))
   
    if lchcolor.lch_h > 0:
        lchcolor.lch_h = (lchcolor.lch_h / math.pi) * 180
    else:
        lchcolor.lch_h = 360 - (math.fabs(lchcolor.lch_h) / math.pi) * 180
    return lchcolor

def Luv_to_XYZ(cobj, debug=False, *args, **kwargs):
    """
    Convert from Luv to XYZ.
    """
    xyzcolor = color_objects.XYZColor()
    _transfer_common(cobj, xyzcolor)
    illum = xyzcolor.get_illuminant_xyz()
   
    # Various variables used throughout the conversion.
    cie_k_times_e = color_constants.CIE_K * color_constants.CIE_E
    u_sub_0 = (4.0 * illum["X"]) / (illum["X"] + 15.0 * illum["Y"] + 3.0 * illum["Z"])
    v_sub_0 = (9.0 * illum["Y"]) / (illum["X"] + 15.0 * illum["Y"] + 3.0 * illum["Z"])
    var_a_frac = (52.0 * cobj.luv_l) / (cobj.luv_u + 13.0 * cobj.luv_l * u_sub_0) 
    var_a = (1.0/3.0) * (var_a_frac - 1.0)
    var_c = -(1.0/3.0)
   
    # Y-coordinate calculations.
    if cobj.luv_l > cie_k_times_e:
        xyzcolor.xyz_y = math.pow((cobj.luv_l + 16.0) / 116.0, 3.0)
    else:
        xyzcolor.xyz_y = cobj.luv_l / color_constants.CIE_K
      
    # These variables depend on Y-coordinate being solved.
    var_b = -5.0 * xyzcolor.xyz_y 
    var_d_frac = (39.0 * cobj.luv_l) / (cobj.luv_v + 13.0 * cobj.luv_l * v_sub_0)
    var_d = xyzcolor.xyz_y * (var_d_frac - 5.0)
   
    # X-coordinate calculation.
    xyzcolor.xyz_x = (var_d - var_b) / (var_a - var_c)
    # Z-coordinate calculation.
    xyzcolor.xyz_z = xyzcolor.xyz_x * var_a + var_b
   
    return xyzcolor

def LCHab_to_Lab(cobj, debug=False, *args, **kwargs):
    """
    Convert from LCH(ab) to Lab.
    """
    labcolor = color_objects.LabColor()
    _transfer_common(cobj, labcolor)
   
    labcolor.lab_l = float(cobj.lch_l)
    labcolor.lab_a = math.cos(math.radians(cobj.lch_h)) * float(cobj.lch_c)
    labcolor.lab_b = math.sin(math.radians(cobj.lch_h)) * float(cobj.lch_c)
    return labcolor

def LCHuv_to_Luv(cobj, debug=False, *args, **kwargs):
    """
    Convert from LCH(uv) to Luv.
    """
    luvcolor = color_objects.LuvColor()
    _transfer_common(cobj, luvcolor)
   
    luvcolor.luv_l = float(cobj.lch_l)
    luvcolor.luv_u = math.cos(math.radians(cobj.lch_h)) * float(cobj.lch_c)
    luvcolor.luv_v = math.sin(math.radians(cobj.lch_h)) * float(cobj.lch_c)
    return luvcolor

def xyY_to_XYZ(cobj, debug=False, *args, **kwargs):
    """
    Convert from xyY to XYZ.
    """
    xyzcolor = color_objects.XYZColor()
    _transfer_common(cobj, xyzcolor)
   
    xyzcolor.xyz_x = (cobj.xyy_x * cobj.xyy_Y) / (cobj.xyy_y)
    xyzcolor.xyz_y = cobj.xyy_Y
    xyzcolor.xyz_z = ((1.0 - cobj.xyy_x - cobj.xyy_y) * xyzcolor.xyz_y) / (cobj.xyy_y)
    
    return xyzcolor

def XYZ_to_xyY(cobj, debug=False, *args, **kwargs):
    """
    Convert from XYZ to xyY.
    """
    xyycolor = color_objects.xyYColor()
    _transfer_common(cobj, xyycolor)
   
    xyycolor.xyy_x = (cobj.xyz_x) / (cobj.xyz_x + cobj.xyz_y + cobj.xyz_z)
    xyycolor.xyy_y = (cobj.xyz_y) / (cobj.xyz_x + cobj.xyz_y + cobj.xyz_z)
    xyycolor.xyy_Y = cobj.xyz_y

    return xyycolor

def XYZ_to_Luv(cobj, debug=False, *args, **kwargs):
    """
    Convert from XYZ to Luv
    """
    luvcolor = color_objects.LuvColor()
    _transfer_common(cobj, luvcolor)
   
    temp_x = cobj.xyz_x
    temp_y = cobj.xyz_y
    temp_z = cobj.xyz_z
   
    luvcolor.luv_u = (4.0 * temp_x) / (temp_x + (15.0 * temp_y) + (3.0 * temp_z))
    luvcolor.luv_v = (9.0 * temp_y) / (temp_x + (15.0 * temp_y) + (3.0 * temp_z))

    illum = luvcolor.get_illuminant_xyz()  
    temp_y = temp_y / illum["Y"]
    if temp_y > color_constants.CIE_E:
        temp_y = math.pow(temp_y, (1.0 / 3.0))
    else:
        temp_y = (7.787 * temp_y) + (16.0 / 116.0)
   
    ref_U = (4.0 * illum["X"]) / (illum["X"] + (15.0 * illum["Y"]) + (3.0 * illum["Z"]))
    ref_V = (9.0 * illum["Y"]) / (illum["X"] + (15.0 * illum["Y"]) + (3.0 * illum["Z"]))
   
    luvcolor.luv_l = (116.0 * temp_y) - 16.0
    luvcolor.luv_u = 13.0 * luvcolor.luv_l * (luvcolor.luv_u - ref_U)
    luvcolor.luv_v = 13.0 * luvcolor.luv_l * (luvcolor.luv_v - ref_V)
   
    return luvcolor

def XYZ_to_Lab(cobj, debug=False, *args, **kwargs):
    """
    Converts XYZ to Lab.
    """
    illum = cobj.get_illuminant_xyz()
    labcolor = color_objects.LabColor()
    _transfer_common(cobj, labcolor)
   
    temp_x = cobj.xyz_x / illum["X"]
    temp_y = cobj.xyz_y / illum["Y"]
    temp_z = cobj.xyz_z / illum["Z"]
   
    if temp_x > color_constants.CIE_E:
        temp_x = math.pow(temp_x, (1.0 / 3.0))
    else:
        temp_x = (7.787 * temp_x) + (16.0 / 116.0)     

    if temp_y > color_constants.CIE_E:
        temp_y = math.pow(temp_y, (1.0 / 3.0))
    else:
        temp_y = (7.787 * temp_y) + (16.0 / 116.0)
   
    if temp_z > color_constants.CIE_E:
        temp_z = math.pow(temp_z, (1.0 / 3.0))
    else:
        temp_z = (7.787 * temp_z) + (16.0 / 116.0)
      
    labcolor.lab_l = (116.0 * temp_y) - 16.0
    labcolor.lab_a = 500.0 * (temp_x - temp_y)
    labcolor.lab_b = 200.0 * (temp_y - temp_z)
    return labcolor

def __downscale_rgb_vals(r, g, b):
    """
    Scales an RGB color object from an int 0-255 to decimal 0.0-1.0.
    """
    var_r = r / 255.0
    var_g = g / 255.0
    var_b = b / 255.0
    
    return (var_r, var_g, var_b)

def __upscale_rgb(rgbcolor):
    """
    Scales an RGB color object from decimal 0.0-1.0 to int 0-255.
    """
    # RGB values are to not go under 0.
    if rgbcolor.rgb_r < 0:
        rgbcolor.rgb_r = 0
    if rgbcolor.rgb_g < 0:
        rgbcolor.rgb_g = 0
    if rgbcolor.rgb_b < 0:
        rgbcolor.rgb_b = 0
      
    # Scale up to 0-255 values.
    rgbcolor.rgb_r = int(math.floor(0.5 + rgbcolor.rgb_r * 255))
    rgbcolor.rgb_g = int(math.floor(0.5 + rgbcolor.rgb_g * 255))
    rgbcolor.rgb_b = int(math.floor(0.5 + rgbcolor.rgb_b * 255))
   
    # Cap RGB values at 255. This shouldn't happen, but it's here just in case
    # things go out of gamut or other fun things.
    if rgbcolor.rgb_r > 255:
        rgbcolor.rgb_r = 255
    if rgbcolor.rgb_g > 255:
        rgbcolor.rgb_g = 255
    if rgbcolor.rgb_b > 255:
        rgbcolor.rgb_b = 255
        
    return rgbcolor

def XYZ_to_RGB(cobj, target_rgb="sRGB", debug=False, *args, **kwargs):
    """
    XYZ to RGB conversion.
    """
    target_rgb = target_rgb.lower()
    rgbcolor = color_objects.RGBColor()
    _transfer_common(cobj, rgbcolor)
    
    temp_X = cobj.xyz_x
    temp_Y = cobj.xyz_y
    temp_Z = cobj.xyz_z
   
    if debug:
        print "  \- Target RGB space: %s" % target_rgb
    target_illum = color_constants.RGB_SPECS[target_rgb]["native_illum"]
    cobj.illuminant = cobj.illuminant.lower()
    if debug:
        print "  \- Target native illuminant: %s" % target_illum
        print "  \- XYZ color's illuminant: %s" % cobj.illuminant
   
    # If the XYZ values were taken with a different reference white than the
    # native reference white of the target RGB space, a transformation matrix
    # must be applied.
    if cobj.illuminant != target_illum:
        if debug:
            print "  \* Applying transformation from %s to %s " % (cobj.illuminant,
                                                                target_illum)
        # Get the adjusted XYZ values, adapted for the target illuminant.
        temp_X, temp_Y, temp_Z = apply_XYZ_transformation(temp_X, temp_Y, temp_Z, 
                                                        orig_illum=cobj.illuminant, 
                                                        targ_illum=target_illum,
                                                        debug=debug)
   
    # Apply an RGB working space matrix to the XYZ values (matrix mul).
    rgbcolor.rgb_r, rgbcolor.rgb_g, rgbcolor.rgb_b = apply_RGB_matrix(temp_X, 
                                          temp_Y, temp_Z, rgb_type=target_rgb, 
                                          convtype="xyz_to_rgb", debug=debug)

    if target_rgb == "srgb":
        # If it's sRGB...
        if rgbcolor.rgb_r > 0.0031308:
            rgbcolor.rgb_r = (1.055 * math.pow(rgbcolor.rgb_r, 1.0 / 2.4)) - 0.055
        else:
            rgbcolor.rgb_r = rgbcolor.rgb_r * 12.92
   
        if rgbcolor.rgb_g > 0.0031308:
            rgbcolor.rgb_g = (1.055 * math.pow(rgbcolor.rgb_g, 1.0 / 2.4)) - 0.055
        else:
            rgbcolor.rgb_g = rgbcolor.rgb_g * 12.92
   
        if rgbcolor.rgb_b > 0.0031308:
            rgbcolor.rgb_b = (1.055 * math.pow(rgbcolor.rgb_b, 1.0 / 2.4)) - 0.055
        else:
            rgbcolor.rgb_b = rgbcolor.rgb_b * 12.92
    else:
        # If it's not sRGB...
        gamma = color_constants.RGB_SPECS[target_rgb]["gamma"]
      
        if rgbcolor.rgb_r < 0:
            rgbcolor.rgb_r = 0
        if rgbcolor.rgb_g < 0:
            rgbcolor.rgb_g = 0
        if rgbcolor.rgb_b < 0:
            rgbcolor.rgb_b = 0
      
        #print "RGB", rgbcolor.rgb_r, rgbcolor.rgb_g, rgbcolor.rgb_b
        rgbcolor.rgb_r = math.pow(rgbcolor.rgb_r, (1 / gamma))
        rgbcolor.rgb_g = math.pow(rgbcolor.rgb_g, (1 / gamma))
        rgbcolor.rgb_b = math.pow(rgbcolor.rgb_b, (1 / gamma))
      
    rgbcolor.rgb_type = target_rgb
    return __upscale_rgb(rgbcolor)

def RGB_to_XYZ(cobj, target_illuminant=None, debug=False, *args, **kwargs):
    """
    RGB to XYZ conversion. Expects 0-255 RGB values.
    """
    xyzcolor = color_objects.XYZColor()
    _transfer_common(cobj, xyzcolor)
    
    temp_R, temp_G, temp_B = __downscale_rgb_vals(cobj.rgb_r,
                                                  cobj.rgb_g,
                                                  cobj.rgb_b)
   
    if cobj.rgb_type == "srgb":
        # If it's sRGB...
        if temp_R > 0.04045:
            temp_R = math.pow((temp_R + 0.055) / 1.055, 2.4)
        else:
            temp_R = temp_R / 12.92
   
        if temp_G > 0.04045:
            temp_G = math.pow((temp_G + 0.055) / 1.055, 2.4)
        else:
            temp_G = temp_G / 12.92
   
        if temp_B > 0.04045:
            temp_B = math.pow((temp_B + 0.055) / 1.055, 2.4)
        else:
            temp_B = temp_B / 12.92
    else:
        # If it's not sRGB...
        gamma = color_constants.RGB_SPECS[cobj.rgb_type]["gamma"]
            
        temp_R = math.pow(temp_R, gamma)
        temp_G = math.pow(temp_G, gamma)
        temp_B = math.pow(temp_B, gamma)
        
    # Apply an RGB working space matrix to the XYZ values (matrix mul).
    xyzcolor.xyz_x, xyzcolor.xyz_y, xyzcolor.xyz_z = apply_RGB_matrix(temp_R, 
                                          temp_G, temp_B, rgb_type=cobj.rgb_type, 
                                          convtype="rgb_to_xyz", debug=debug)
    
    if target_illuminant == None:
        target_illuminant = color_constants.RGB_SPECS[cobj.rgb_type]["native_illum"]
        
    # The illuminant of the original RGB object.
    source_illuminant = color_constants.RGB_SPECS[cobj.rgb_type]["native_illum"]
    
    # This needs to be correct before the adaptation is applied.
    xyzcolor.illuminant = source_illuminant
    # This will take care of any illuminant changes for us.
    xyzcolor.apply_adaptation(target_illuminant)

    return xyzcolor

def __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max):
    """
    For RGB_to_HSL and RGB_to_HSV, the Hue (H) component is calculated in
    the same way.
    """
    if var_max == var_min:
        return 0.0
    elif var_max == var_R:
        return (60.0 * ((var_G - var_B) / (var_max - var_min)) + 360) % 360.0
    elif var_max == var_G:
        return 60.0 * ((var_B - var_R) / (var_max - var_min)) + 120
    elif var_max == var_B:
        return 60.0 * ((var_R - var_G) / (var_max - var_min)) + 240.0
    
def RGB_to_HSV(cobj, debug=False, *args, **kwargs):
    """
    Converts from RGB to HSV.
    
    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    V values are a percentage, 0.0 to 1.0.
    """
    hsvcolor = color_objects.HSVColor()
    _transfer_common(cobj, hsvcolor)
    
    var_R = cobj.rgb_r / 255.0
    var_G = cobj.rgb_g / 255.0
    var_B = cobj.rgb_b / 255.0
    
    var_max = max(var_R, var_G, var_B)
    var_min = min(var_R, var_G, var_B)
    
    var_H = __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max)
    
    if var_max == 0:
        var_S = 0
    else:
        var_S = 1.0 - (var_min / var_max)
        
    var_V = var_max
    
    hsvcolor.rgb_type = cobj.rgb_type    
    hsvcolor.hsv_h = var_H
    hsvcolor.hsv_s = var_S
    hsvcolor.hsv_v = var_V

    return hsvcolor

def RGB_to_HSL(cobj, debug=False, *args, **kwargs):
    """
    Converts from RGB to HSL.
    
    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    L values are a percentage, 0.0 to 1.0.
    """
    hslcolor = color_objects.HSLColor()
    _transfer_common(cobj, hslcolor)
    
    var_R = cobj.rgb_r / 255.0
    var_G = cobj.rgb_g / 255.0
    var_B = cobj.rgb_b / 255.0
    
    var_max = max(var_R, var_G, var_B)
    var_min = min(var_R, var_G, var_B)
    
    var_H = __RGB_to_Hue(var_R, var_G, var_B, var_min, var_max)
    var_L = 0.5 * (var_max + var_min)
    
    if var_max == var_min:
        var_S = 0
    elif var_L <= 0.5:
        var_S = (var_max - var_min) / (2.0 * var_L)
    else:
        var_S = (var_max - var_min) / (2.0 - (2.0 * var_L))
    
    hslcolor.rgb_type = cobj.rgb_type
    hslcolor.hsl_h = var_H
    hslcolor.hsl_s = var_S
    hslcolor.hsl_l = var_L
    
    return hslcolor

def __Calc_HSL_to_RGB_Components(var_q, var_p, C):
    """
    This is used in HSL_to_RGB conversions on R, G, and B.
    """
    if C < 0:
        C += 1.0
    if C > 1:
        C -= 1.0

    # Computing C of vector (Color R, Color G, Color B)
    if C < (1.0 / 6.0):
        return var_p + ((var_q - var_p) * 6.0 * C)
    elif (1.0 / 6.0) <= C < 0.5:
        return var_q
    elif 0.5 <= C < (2.0 / 3.0):
        return var_p + ((var_q - var_p) * 6.0 * ((2.0 / 3.0) - C))
    else:
        return var_p
    
def HSV_to_RGB(cobj, target_rgb=None, debug=False, *args, **kwargs):
    """
    HSV to RGB conversion.
    
    H values are in degrees and are 0 to 360.
    S values are a percentage, 0.0 to 1.0.
    V values are a percentage, 0.0 to 1.0.
    """
    rgbcolor = color_objects.RGBColor()
    _transfer_common(cobj, rgbcolor)
    
    H = cobj.hsv_h
    S = cobj.hsv_s
    V = cobj.hsv_v
    
    h_floored = int(math.floor(H))
    h_sub_i = int(h_floored / 60) % 6
    var_f = (H / 60.0) - (h_floored / 60)
    var_p = V * (1.0 - S)
    var_q = V * (1.0 - var_f * S)
    var_t = V * (1.0 - (1.0 - var_f) * S)
       
    if h_sub_i == 0:
        rgbcolor.rgb_r = V
        rgbcolor.rgb_g = var_t
        rgbcolor.rgb_b = var_p
    elif h_sub_i == 1:
        rgbcolor.rgb_r = var_q
        rgbcolor.rgb_g = V
        rgbcolor.rgb_b = var_p
    elif h_sub_i == 2:
        rgbcolor.rgb_r = var_p
        rgbcolor.rgb_g = V
        rgbcolor.rgb_b = var_t
    elif h_sub_i == 3:
        rgbcolor.rgb_r = var_p
        rgbcolor.rgb_g = var_q
        rgbcolor.rgb_b = V
    elif h_sub_i == 4:
        rgbcolor.rgb_r = var_t
        rgbcolor.rgb_g = var_p
        rgbcolor.rgb_b = V
    elif h_sub_i == 5:
        rgbcolor.rgb_r = V
        rgbcolor.rgb_g = var_p
        rgbcolor.rgb_b = var_q
    
    __upscale_rgb(rgbcolor)
    # In the event that they define an HSV color and want to convert it to 
    # a particular RGB space, let them override it here.
    if target_rgb != None:
        rgbcolor.rgb_type = target_rgb
    else:
        rgbcolor.rgb_type = cobj.rgb_type
        
    return rgbcolor

def HSL_to_RGB(cobj, target_rgb=None, debug=False, *args, **kwargs):
    """
    HSL to RGB conversion.
    """
    rgbcolor = color_objects.RGBColor()
    _transfer_common(cobj, rgbcolor)
    
    H = cobj.hsl_h
    S = cobj.hsl_s
    L = cobj.hsl_l
    
    if L < 0.5:
        var_q = L * (1.0 + S)
    else:
        var_q = L + S - (L * S)
        
    var_p = 2.0 * L - var_q
    
    # H normalized to range [0,1]
    h_sub_k = (H / 360.0)
    
    t_sub_R = h_sub_k + (1.0 / 3.0)
    t_sub_G = h_sub_k
    t_sub_B = h_sub_k - (1.0 / 3.0)
    
    rgbcolor.rgb_r = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_R)
    rgbcolor.rgb_g = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_G)
    rgbcolor.rgb_b = __Calc_HSL_to_RGB_Components(var_q, var_p, t_sub_B)

    __upscale_rgb(rgbcolor)
    # In the event that they define an HSV color and want to convert it to 
    # a particular RGB space, let them override it here.
    if target_rgb != None:
        rgbcolor.rgb_type = target_rgb
    else:
        rgbcolor.rgb_type = cobj.rgb_type    
    
    return rgbcolor

def RGB_to_CMY(cobj, debug=False, *args, **kwargs):
    """
    RGB to CMY conversion.
    
    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """
    cmycolor = color_objects.CMYColor()
    _transfer_common(cobj, cmycolor)
   
    cmycolor.cmy_c = 1.0 - (cobj.rgb_r / 255.0)
    cmycolor.cmy_m = 1.0 - (cobj.rgb_g / 255.0)
    cmycolor.cmy_y = 1.0 - (cobj.rgb_b / 255.0)
    
    return cmycolor

def CMY_to_RGB(cobj, debug=False, *args, **kwargs):
    """
    Converts CMY to RGB via simple subtraction.
    
    NOTE: Returned values are in the range of 0-255.
    """
    rgbcolor = color_objects.RGBColor()
    _transfer_common(cobj, rgbcolor)
    
    rgbcolor.rgb_r = 1.0 - cobj.cmy_c
    rgbcolor.rgb_g = 1.0 - cobj.cmy_m
    rgbcolor.rgb_b = 1.0 - cobj.cmy_y
    
    return __upscale_rgb(rgbcolor)

def CMY_to_CMYK(cobj, debug=False, *args, **kwargs):
    """
    Converts from CMY to CMYK.
    
    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """ 
    cmykcolor = color_objects.CMYKColor()
    _transfer_common(cobj, cmykcolor)
   
    var_k = 1.0
    if cobj.cmy_c < var_k:
        var_k = cobj.cmy_c
    if cobj.cmy_m < var_k:
        var_k = cobj.cmy_m
    if cobj.cmy_y < var_k:
        var_k = cobj.cmy_y
      
    if var_k == 1:
        cmykcolor.cmyk_c = 0.0
        cmykcolor.cmyk_m = 0.0
        cmykcolor.cmyk_y = 0.0
    else:
        cmykcolor.cmyk_c = (cobj.cmy_c - var_k) / (1.0 - var_k)
        cmykcolor.cmyk_m = (cobj.cmy_m - var_k) / (1.0 - var_k)
        cmykcolor.cmyk_y = (cobj.cmy_y - var_k) / (1.0 - var_k)
    cmykcolor.cmyk_k = var_k

    return cmykcolor

def CMYK_to_CMY(cobj, debug=False, *args, **kwargs):
    """
    Converts CMYK to CMY.
    
    NOTE: CMYK and CMY values range from 0.0 to 1.0
    """
    cmycolor = color_objects.CMYColor()
    _transfer_common(cobj, cmycolor)
    
    cmycolor.cmy_c = cobj.cmyk_c * (1.0 - cobj.cmyk_k) + cobj.cmyk_k
    cmycolor.cmy_m = cobj.cmyk_m * (1.0 - cobj.cmyk_k) + cobj.cmyk_k
    cmycolor.cmy_y = cobj.cmyk_y * (1.0 - cobj.cmyk_k) + cobj.cmyk_k
    
    return cmycolor

import color_objects


"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module contains exceptions for use throughout the L11 Colorlib.
"""
class BaseException(Exception):
    def __str__(self):
        return repr(self.value)

class MissingValue(BaseException):
    """
    Raised when one of the required components (Lab, LCH, etc) is missing.
    """
    def __init__(self, cobj, missing_val_name):
        self.value = "Missing %s value on %s." % (missing_val_name, 
                                                  cobj.__class__.__name__)
        
class InvalidValue(BaseException):
    """
    Raised when one of the required components is invalid.
    """
    def __init__(self, cobj, invalid_val_name, invalid_val):
        self.value = "Invalid %s value (%s) on %s." % (invalid_val_name,
                                                       invalid_val, 
                                                       cobj.__class__.__name__)
        
class InvalidDeltaEMode(BaseException):
    """
    Raised when an invalid Delta E mode is specified.
    """
    def __init__(self, mode):
        self.value = "Invalid Delta E mode: %s" % mode
        
class InvalidArgument(BaseException):
    """
    Raised when an invalid argument is passed to a function.
    """
    def __init__(self, function_name, invalid_arg_name, invalid_arg):
        self.value = "Invalid argument for %s (%s) passed to %s()." % (
                                                            invalid_arg_name,
                                                            invalid_arg,
                                                            function_name)
    
class InvalidConversion(BaseException):
    """
    Raised when the user asks for a color space conversion that does not exist.
    """
    def __init__(self, cobj, cs_to):
        self.value = "Conversion from %s to %s is not defined." % (cobj, cs_to)
    
class InvalidIlluminant(BaseException):
    """
    Raised when an invalid illuminant is set on a ColorObj.
    """
    def __init__(self, cobj):
        self.value = "Invalid illuminant specified: %s" % (cobj.illuminant)
    
class InvalidObserver(BaseException):
    """
    Raised when an invalid observer is set on a ColorObj.
    """
    def __init__(self, cobj):
        self.value = "Invalid observer angle specified: %s" % (cobj.observer)


"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This module shows some examples of Delta E calculations of varying types.
"""
import example_config
from colormath.color_objects import *

# Reference color.
color1 = LabColor(lab_l=0.9, lab_a=16.3, lab_b=-2.22)
# Color to be compared to the reference.
color2 = LabColor(lab_l=0.7, lab_a=14.2, lab_b=-1.80)

print "== Delta E Colors =="
print " COLOR1: %s" % color1
print " COLOR2: %s" % color2
print "== Results =="
print " CIE1976: %.3f" % color1.delta_e(color2, mode='cie1976')
print " CIE1994: %.3f (Graphic Arts)" % color1.delta_e(color2, mode='cie1994')
# Different values for textiles.
print " CIE1994: %.3f (Textiles)" % color1.delta_e(color2, mode='cie1994', 
                                                   K_1=0.048, K_2=0.014, 
                                                   K_L=2)
print " CIE2000: %.3f" % color1.delta_e(color2, mode='cie2000')
# Typically used for acceptability.
print "     CMC: %.3f (2:1)" % color1.delta_e(color2, mode='cmc', p1=2, pc=1)
# Typically used to more closely model human percetion.
print "     CMC: %.3f (1:1)" % color1.delta_e(color2, mode='cmc', p1=1, pc=1)


#!/usr/bin/env python
"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from distutils.core import setup
import colormath

LONG_DESCRIPTION = \
"""Implements a large number of different color operations such as
color space conversions, Delta E, and density to spectral."""

CLASSIFIERS = [
                'Development Status :: 5 - Production/Stable',
                'Intended Audience :: Developers',
                'License :: OSI Approved :: GNU General Public License (GPL)',
                'Natural Language :: English',
                'Operating System :: OS Independent',
                'Programming Language :: Python',
                'Topic :: Scientific/Engineering :: Mathematics',
                'Topic :: Software Development :: Libraries :: Python Modules' 
              ]

KEYWORDS = 'color math conversions'

setup(name='colormath',
      version=colormath.VERSION,
      description='Color math and conversion library.',
      long_description = LONG_DESCRIPTION,
      author='Gregory Taylor',
      author_email='gtaylor@l11solutions.com',
      url='http://code.google.com/p/python-colormath/',
      download_url='http://pypi.python.org/pypi/colormath/',
      packages=['colormath'],
      platforms = ['Platform Independent'],
      license = 'GPLv3',
      classifiers = CLASSIFIERS,
      keywords = KEYWORDS,
      requires = ['numpy']
     )



"""
 Color Math Module (colormath) 
 Copyright (C) 2009 Gregory Taylor

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
Formulas for density calculation.
"""
from math import log10, log
from colormath.density_standards import *

def ansi_density(color, std_array):
    """
    Calculates density for the given SpectralColor using the spectral weighting
    function provided. For example, ANSI_STATUS_T_RED. These may be found in
    density_standards.py.
    
    color: (SpectralColor) The SpectralColor object to calculate density for.
    std_array: (ndarray) NumPy array of filter of choice from density_standards.py.
    """  
    # Load the spec_XXXnm attributes into a Numpy array.
    sample = color.get_numpy_array()
    # Matrix multiplication
    intermediate = sample * std_array
    
    # Sum the products.
    numerator = intermediate.sum()
    # This is the denominator in the density equation.
    sum_of_standard_wavelengths = std_array.sum()
    
    # This is the top level of the density formula.
    return -1.0 * log10(numerator / sum_of_standard_wavelengths)

def auto_density(color):
    """
    Given a SpectralColor, automatically choose the correct ANSIT filter. Returns
    a tuple with a string representation of the filter the calculated density.
    """
    color_array = color.get_numpy_array()
    blue_density = ansi_density(color, ANSI_STATUS_T_BLUE)
    green_density = ansi_density(color, ANSI_STATUS_T_GREEN)
    red_density = ansi_density(color, ANSI_STATUS_T_RED)
    
    densities = [blue_density, green_density, red_density]
    min_density = min(densities)
    max_density = max(densities)
    density_range = max_density - min_density
    #print "DIFF", density_range
    
    # See comments in density_standards.py for VISUAL_DENSITY_THRESH to
    # understand what this is doing.
    if density_range <= VISUAL_DENSITY_THRESH:
        #print "visual"
        return ansi_density(color, ISO_VISUAL)
    elif blue_density > green_density and blue_density > red_density:
        #print "blue"
        return blue_density
    elif green_density > blue_density and green_density > red_density:
        #print "green"
        return green_density
    else:
        #print "red"
        return red_density


"""
This file holds various configuration options used for all of the examples.
"""
import os
import sys
# Use the colormath directory included in the downloaded package instead of
# any globally installed versions.
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))